"use strict";
/*
    This file is part of the Pale Moon Locale Switcher extension.
    
    Copyright (C) 2014,2015 Jeremy Morton
    Copyright (C) 2016 Mark Straver BASc

    The Pale Moon Locale Switcher extension is free software: you can
    redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either
    version 3 of the License, or (at your option) any later version.

    The Pale Moon Locale Switcher extension is distributed in the hope that
    it will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
// Javascript Code Module - Dynamic UI
// Provides functionality for dynamically modifying the user interface of the XUL-based application being worked with.

var EXPORTED_SYMBOLS = ["dynamicUI"];

var dynamicUI = (function(my, moduleGlobalScope) {
	
	// Private members (please don't use them directly from calling code)
	// ==================================================================
	
	my._namespaces = {
		html: "http://www.w3.org/1999/xhtml",
		xul: "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"
	};
	my._defaultNamespace = my._namespaces.xul;
	
	
	
	// Public members
	// ==============
	
	// Converts a JSON template to the DOM structure it represents.  Defaults to putting each non-prefixed XML element into the XUL namespace.
	// 
	// jsonTemplate: A correctly-formatted JSON template representing the structure and content of the DOM nodes to be created.  Each node should be an
	//               array with 3 entries; element name (or array of elements), the element's attributes, and the element's child elements.  Example JSON
	//               template that is correctly-formatted and that may be passed into this function:
	//	[ "hbox",
	//		{},
	//		[ "html:div",
	//			{},
	//			[ "html:a",
	//				{
	//					href: href,
	//					key: "link",
	//					onclick: function(event) { alert(event.target.href); }
	//				},
	//				"link text"
	//			],
	//			[ "html:br",
	//				{},
	//			  // !! IF A NODE SHOULD HAVE NO CHILD NODES, pass null for its third array entry (an empty array is INCORRECT) !!
	//				null
	//			],
	//			  // Element 'fullName' (ie. with optional namespace prefix)
	//			[ "html:span",
	//			  // Element attributes
	//				{
	//					class: "stuff"
	//				},
	//			  // Child element(s) - string for a text node, array for a normal DOM node specification, or an actual DOM node to insert
	//				"stuff"
	//			]
	//		]
	//	]
	// doc: The document into which the DOM structure will be inserted.
	// nodes: An empty object that will be populated with any DOM nodes that have the "tag" attribute set.  For example, in the above jsonTemplate example,
	//        the A node with |key: "link"| will result in 'nodes' containing a property named 'link', which will be a reference to the created A node.
	// 
	// returns: The DOM element (or array of elements, if an array was passed as the first jsonTemplate argument) representing the JSON template passed in
	//          to the function.  This can then be safely appended to the document passed in via the 'doc' argument.
	my.JsonToDOM = function(jsonTemplate, doc, nodes) {
		function namespace(name) {
			var reElemNameParts = /^(?:(.*):)?(.*)$/.exec(name);
			return { namespace: my._namespaces[reElemNameParts[1]], shortName: reElemNameParts[2] };
		};
		
		// Note that 'elemNameOrArray' is: either the full element name (eg. [html:]div) or an array of elements in JSON notation
		function tag(elemNameOrArray, elemAttr) {
			// Array of elements?  Parse each one...
			if (Array.isArray(elemNameOrArray)) {
				var frag = doc.createDocumentFragment();
				Array.forEach(arguments, function(thisElem) {
					frag.appendChild(tag.apply(null, thisElem));
				});
				return frag;
			}
			
			// Single element?  Parse element namespace prefix (if none exists, default to defaultNamespace), and create element
			var elemNs = namespace(elemNameOrArray);
			var elem = doc.createElementNS(elemNs.namespace || my._defaultNamespace, elemNs.shortName);
			// ^ note that AMO warns about this as we're using a variable with createElementNS, but it's a silly warning, and this code is fine...
			
			// Set element's attributes and/or callback functions (eg. onclick)
			for (var key in elemAttr) {
				var val = elemAttr[key];
				if (nodes && key == "key") {
					nodes[val] = elem;
					continue;
				}
				
				var attrNs = namespace(key);
				if (typeof val == "function") {
					// Special case for function attributes; don't just add them as 'on...' attributes, but as events, using addEventListener
					elem.addEventListener(key.replace(/^on/, ""), val, false);
				}
				else {
					// Note that the default namespace for XML attributes is, and should be, blank (ie. they're not in any namespace)
					elem.setAttributeNS(attrNs.namespace || "", attrNs.shortName, val);
				}
			}
			
			// Create and append this element's children
			var childElems = Array.slice(arguments, 2);
			childElems.forEach(function(childElem) {
				if (childElem != null) {
					elem.appendChild(
						typeof childElem == "object" ? tag.apply(null, childElem) :
							childElem instanceof Node ? childElem :
								doc.createTextNode(childElem)
					);
				}
			});
			
			return elem;
		};
		
		// 'apply' the template (which is a JS array) to the 'tag' function; 'apply' takes 2 arguments; 'thisArg' (what 'this' should be during the context
		// of the function call, which will be 'the global object' if null), and 'argsArray' (an array of arguments to pass to the function) - then return
		// the result, which will be the DOM node(s) structure generated from the JSON template.
		return tag.apply(null, jsonTemplate);
	};
	
	// Searches the toolbox's toolbars for an item with a given ID, and removes it if it's found.
	// 
	// toolbox: The toolbox whose toolbars to search for the item to remove.
	// itemId: The ID of the item to remove.
	// 
	// returns: If the item was found and removed, the item's DOM object; otherwise null.
	my.RemoveToolbarItem = function(toolbox, itemId) {
		for (let i=0; i < toolbox.childNodes.length; i++) {
			let thisToolbar = toolbox.childNodes[i];
			if (thisToolbar.localName == "toolbar") {
				let retVal = my.RemoveChildNodeItem(thisToolbar, itemId);
				if (retVal) { return retVal; }
			}
		}
		
		return null;
	};
	
	// Searches a JS object with child DOM nodes for a node with a given ID, and removes it if it's found.
	// 
	// nodesParent: The parent node that contains the child nodes to search through.
	// nodeId: The ID of the child node to remove.
	// 
	// returns: If the node was found and removed, the node's DOM object; otherwise null.
	my.RemoveChildNodeItem = function(nodesParent, nodeId) {
		for (let i=0; i < nodesParent.childNodes.length; i++) {
			let thisChildItem = nodesParent.childNodes[i];
			if (thisChildItem.id == nodeId) {
				// Found node to remove.  Remove it, and return it.
				return nodesParent.removeChild(thisChildItem);
			}
		}
	};
	
	// *** Finally ***
	// Return |my| - everything we've assigned to it will be exported to the public var we're defining.
	return my;
	
}((dynamicUI || {}), this));
