"use strict";
/*
    This file is part of the Pale Moon Locale Switcher extension.
    
    Copyright (C) 2014,2015 Jeremy Morton
    Copyright (C) 2016 Mark Straver BASc

    The Pale Moon Locale Switcher extension is free software: you can
    redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either
    version 3 of the License, or (at your option) any later version.

    The Pale Moon Locale Switcher extension is distributed in the hope that
    it will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
// Javascript Code Module - Miscellaneous utilities v2.7
// Containins miscellaneous utilities for use with extension code.  Also has the ability to persistently store data across different
// contexts, via the DataStore feature (the persistence works specifically because this is a JS Code Module, and therefore a singleton).
// Relies on being able to pull in extension-specific info from "extInfo.jsm" JS module.

var EXPORTED_SYMBOLS = ["miscUtils"];

// Components shortcuts
const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;

var miscUtils = (function(my, moduleGlobalScope) {
	
	// *** Import JSMs into 'jsm' object ***
	my.jsm = my.jsm || {};
	// XPCOM utilities, exports symbol(s) [ "XPCOMUtils" ]
	Components.utils.import("resource://gre/modules/XPCOMUtils.jsm", my.jsm);
	// Necko utilities, exports symbol(s) [ "NetUtil" ]
	Components.utils.import("resource://gre/modules/NetUtil.jsm", my.jsm);
	// File utilities, exports symbol(s) [ "FileUtils" ]
	Components.utils.import("resource://gre/modules/FileUtils.jsm", my.jsm);
	// Extension-specific info provider, exports symbol(s) [ "extInfo" ]
	my.jsm.XPCOMUtils.importRelative(moduleGlobalScope, "extInfo.jsm", my.jsm);
	
	// Private members (please don't use them directly from calling code)
	// ==================================================================
	
	my._appName = null;
	my._appPlatform = null;
	my._appVer = null;
	my._appVerMajorOnly = null;
	my._strBundleService = null;
	my._dataStore = new Object();
	my._rootPrefBranch = null;
	my._versionChecker = null;
	
	
	
	// Public members
	// ==============
	
	// Gets the XUL namespace as a string
	// 
	// returns: the XUL namespace as a string.
	my.GetXulNamespace = function() {
		return "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
	};
	
	// Dump debug output to console window (on Windows, must start Firefox/SeaMonkey/etc. with -console argument to see output).
	// 
	// dumpText: The text to output.
	// lineSeparator: (optional) The string to use at the end of the text.  Defaults to a single newline.
	// 
	// returns: null.
	my.Dump = function(dumpText, lineSeparator) {
		var lineSep = '\n';
		if (lineSeparator) { lineSep = lineSeparator; }
		
		dump(my.jsm.extInfo.GetConciseDisplayName() + " | " + dumpText + lineSep);
	};
	
	// Return the name of the application we're currently running on.
	// 
	// returnUnrecognized: (optional) If true, returns the application name string when the application name is unrecognized; otherwise, returns null when
	// it's unrecognized.
	// 
	// returns: A string indicating the recognized or unrecognized name of the application we're currently running on, or potentially null if the
	//          application name is unrecognized.
	my.AppName = function(returnUnrecognized) {
		if (my._appName == null) {
			my._appName =
				Cc["@mozilla.org/xre/app-info;1"]
				.getService(Ci.nsIXULAppInfo)
				.name.toLowerCase();
		}
		
		if (my._appName == "seamonkey") {
			return "recognized:sm";
		}
		else if (my._appName == "firefox") {
			return "recognized:ff";
		}
		else if (my._appName == "pale moon") {
			return "recognized:pm";
		}
		else {
			my.ConsoleOutErr("Unrecognized application name string: " + my._appName, true);
			
			if (returnUnrecognized) { return "unrecognized:" + my._appName; }
			else { return null; }
		}
	};
	
	// Return the application platform/OS we're currently running on.
	// 
	// returnUnrecognized: (optional) If true, returns the platform string when the platform is unrecognized; otherwise, returns null when it's unrecognized.
	// 
	// returns: A string indicating the recognized or unrecognized platform/OS we're currently running on, or potentially null if platform is unrecognized.
	my.AppPlatform = function(returnUnrecognized) {
		if (my._appPlatform == null) {
			my._appPlatform =
				Cc["@mozilla.org/xre/app-info;1"]
				.getService(Ci.nsIXULRuntime)
				.OS;
		}
		
		if (my._appPlatform == "WINNT") {
			return "recognized:win";
		}
		else if (my._appPlatform == "Linux") {
			return "recognized:lin";
		}
		else if (my._appPlatform == "Darwin") {
			return "recognized:mac";
		}
		else {
			my.ConsoleOutErr("Unrecognized application platform/OS string: " + my._appPlatform, true);
			
			if (returnUnrecognized) { return "unrecognized:" + my._appPlatform; }
			else { return null; }
		}
	};
	
	// Return the version of the application we're currently running on.
	// 
	// returnMajorOnly: (optional) If true, returns only the major version number of the full version string.
	// 
	// returns: Version of the application we're currently running on.
	my.AppVersion = function(returnMajorOnly) {
		if (my._appVer == null) {
			// Determine which of the application versions we're running in.  We need a different version output
			// for every different version of the app for which we need to differ our functionality.  The different
			// versions that this function may output with returnMajorOnly set to true are:
			// - 0 (invalid, if we get this we can probably assume things will be screwed up)
			// - 2 (Firefox/SeaMonkey/etc. 2.*)
			// - 3 (Firefox/SeaMonkey/etc. 3.*)
			// - 4 (Firefox/SeaMonkey/etc. 4.*)
			// - etc.
			// 
			// If returnMajorOnly is false, we'll get things like:
			// - 0 (again, invalid)
			// - 2.20 (eg. SeaMonkey 2.20)
			// - 2.22a2 (eg. SeaMonkey 2.22 alpha 2)
			// - 2.23b1 (eg. SeaMonkey 2.23 beta 1)
			// - etc.
			
			// If count gets beyond a huge number and still hasn't determined app version, give up and these 'error' values
			// will be kept.
			my._appVer = 0;
			my._appVerMajorOnly = 0;
			
			var appInfo =
				Cc["@mozilla.org/xre/app-info;1"]
				.getService(Ci.nsIXULAppInfo);
			my._appVer = appInfo.version;
			// We assume that the earliest version of the app that we may be running the extension on is an alpha (eg. "3.0a1").
			for (var count = 2; count < 999999; count++) {
				if ((my.CompareAppVersions(appInfo.version, count + ".0a1") >= 0) && (my.CompareAppVersions(appInfo.version, (count+1) + ".0a1") < 0)) {
					my._appVerMajorOnly = count;
					break;
				}
			}
		}
		
		if (returnMajorOnly) { return my._appVerMajorOnly; }
		else { return my._appVer; }
	};
	
	// Uses the version comparator service to compare two supplied version strings.
	// 
	// verA: The first version string to be compared.
	// verB: The second version string to be compared.
	// 
	// returns: If the two versions are the same, 0.  If verA is lower than verB, below zero.  If verA is higher than verB, above zero.
	my.CompareAppVersions = function(verA, verB) {
		if (!my._versionChecker) {
			try {
				my._versionChecker =
					Cc["@mozilla.org/xpcom/version-comparator;1"]
					.getService(Ci.nsIVersionComparator);
			}
			catch (ex) {
				my.Dump("CompareAppVersions() failed, couldn't get version comparator: " + ex);
				return null;
			}
		}
		
		return my._versionChecker.compare(verA, verB);
	};
	
	// Checks whether the current application version is lower than the given application version.
	// 
	// appVer: The version to compare the version of the current application with.
	// 
	// returns: If the current application version is lower than the given application version, true; otherwise false.
	my.AppVersionBelow = function(appVer) {
		return my.CompareAppVersions(my.AppVersion(), appVer) < 0;
	};
	
	// Get the specified string bundle; originally based on the code that was at chrome://global/content/strres.js
	// 
	// path: Path to the string bundle to get.
	// nullIfEmpty: If true, null is returned if the string bundle's enumerated strings list has no entries (or the bundle isn't chrome-registered at all).
	// 
	// returns: The specified string bundle, or null if there was an error (or nullIfEmpty & empty strings list / bundle isn't chrome-registered).
	my.GetStrBundle = function(path, nullIfEmpty) {
		var strBundle = null;
		
		if (!my._strBundleService) {
			try {
				my._strBundleService =
					Cc["@mozilla.org/intl/stringbundle;1"]
					.getService(Ci.nsIStringBundleService);
			}
			catch (ex) {
				my.Dump("GetStrBundle() failed, couldn't get string bundle service: " + ex);
				return null;
			}
		}
		
		strBundle = my._strBundleService.createBundle(path);
		if (!strBundle) {
			my.Dump("GetStrBundle() failed, couldn't get string bundle: " + ex);
		}
		
		if (strBundle && nullIfEmpty) {
			var ioService =
				Cc["@mozilla.org/network/io-service;1"]
				.getService(Ci.nsIIOService);
			var chromeRegService =
				Cc["@mozilla.org/chrome/chrome-registry;1"]
				.getService(Ci.nsIToolkitChromeRegistry);
			
			var bundleUri = ioService.newURI(path, null, null);
			
			if (!chromeRegService.getLocalesForPackage(bundleUri.host).hasMore()) { strBundle = null; }
			else {
				// We do have a chrome locale provider registered for this packagename, so try to access the bundle and enumerate through its elements
				try {
					var count = 0;
					var strings = strBundle.getSimpleEnumeration();
					while (strings.hasMoreElements()) { count++; strings.getNext(); }
					if (count == 0) { strBundle = null; }
				}
				catch (ex) {
					// File probably doesn't exist
					strBundle = null;
				}
			}
		}
		
		return strBundle;
	};
	
	// Outputs the specified string to the 'Error console' (as an informational message)
	// 
	// consoleMsg: The message to output to the console.
	// 
	// returns: null.
	my.ConsoleOut = function(consoleMsg) {
		var consoleService =
			Cc["@mozilla.org/consoleservice;1"]
			.getService(Ci.nsIConsoleService);
		consoleService.logStringMessage(consoleMsg);
	};
	
	// Outputs the specified warning or error to the 'Error console' (as a warning or error).  Can be used to output a simple
	// error which just gives the extension ID, or a full error with file, line number, col number, etc.  If no 'Advanced
	// error information' is passed, will output the 'simple error' style.
	// 
	// errDescription: Error description.
	// isWarning: (optional) If a warning should display rather than an error, true; otherwise false or null.
	// errFile: (optional) Advanced error information; name of file in which error occurred.
	// errLineContents: (optional) Advanced error information; contents of line on which error occurred.
	// errLineNum: (optional) Advanced error information; line number of line contents.
	// errColumnNum: (optional) Advanced error information; column number in which error occurred.
	// 
	// returns: null.
	my.ConsoleOutErr = function(errDescription, isWarning, errFile, errLineContents, errLineNum, errColumnNum) {
		if (
			errFile == null &&
			errLineContents == null &&
			errLineNum == null &&
			errColumnNum == null
		) {
			errFile = "";
			errLineContents = "(in extension " + my.jsm.extInfo.GetExtId() + ")";
			errLineNum = 0;
			errColumnNum = 0;
		}
		
		var consoleService =
			Cc["@mozilla.org/consoleservice;1"]
			.getService(Ci.nsIConsoleService);
		
		var intFlags = 0;
		if (isWarning) { intFlags |= 1; }
		
		var errorObject =
			Cc["@mozilla.org/scripterror;1"]
			.createInstance(Ci.nsIScriptError);
		errorObject.init(errDescription, errFile, errLineContents, errLineNum, errColumnNum, intFlags, "");
		consoleService.logMessage(errorObject);
	};
	
	// Pops up a relatively intrusive Javascript alert dialog with an error message.  Note that extension code should NEVER use
	// the alert() function (or any similar functionality) directly, but should go through this function.  Also, this function
	// itself should be used VERY sparingly - perhaps instead of displaying a dialog if there is a catastrophic error when trying
	// to display the dialog.  The large majority of the time, errors should be indicated in other ways such as through other
	// extension GUI elements, or outputting to console.
	// 
	// errorMsg: The fatal error message to display in the popup.
	// errorTitle: (optional) The title of the popup error message window.
	// parentWindow: (optional) If specified, an nsIDOMWindow specifying the parent window for the alert popup dialog.  Otherwise,
	//                          the parent window will be nsIWindowWatcher.activeWindow.
	// 
	// returns: If fatal error message was successfully displayed, true; otherwise false.
	my.Fatal = function(errorMsg, errorTitle, parentWindow) {
		if (!errorMsg) {
			// No error message given, can't display error.
			my.ConsoleOutErr("Can't display fatal error message popup because no error message was specified.", true);
		}
		if (!errorTitle) {
			errorTitle = "[JavaScript Application]";
		}
		if (!parentWindow) { parentWindow = null; }
		
		// Get the Prompt Service, which is what's generally used for giving modal prompts to the user
		var promptService =
			Cc["@mozilla.org/embedcomp/prompt-service;1"]
			.getService(Ci.nsIPromptService);
		
		var errOutput = "FATAL ERROR: " + my.jsm.extInfo.GetConciseDisplayName() + " | " + errorMsg;
		my.ConsoleOutErr(errOutput);
		promptService.alert(parentWindow, errorTitle, errOutput);
		
		return true;
	};
	
	// Gets the ID of this extension
	// 
	// returns: The ID of this extension.
	my.GetExtId = function() {
		return my.jsm.extInfo.GetExtId();
	};
	
	// Convenience function to return the escaped ID of this extension (escaped by 'EscapeExtId').
	// 
	// returns: The escaped extension ID.
	my.GetEscapedExtId = function() {
		return my.EscapeExtId(my.GetExtId());
	};
	
	// Gets the extension ID passed, escaped into "extensions.extID." format.
	// 
	// extID: The extension ID to escape.
	// 
	// returns: The escaped extension ID.
	my.EscapeExtId = function(extID) {
		var escapedExtID = extID;
		escapedExtID = escapedExtID.replace(/\@/g, '{at}');
		escapedExtID = escapedExtID.replace(/\./g, '{dot}');
		escapedExtID = 'extensions.' + escapedExtID + '.';
		
		return escapedExtID;
	};
	
	// Gets or sets a value from/to the persistent data store key/value table.
	// 
	// key: The key to get or set from/to the persistent data store table.
	// value: (optional) The value to set the given key to in the persistent data store table.  If specified, this function performs
	//                   a 'set' operation and sets the given key to the given value.  Otherwise, it performs a 'get' operation, and
	//                   returns the value of the given key in the persistent data store.
	// 
	// returns: If a get operation is being performed, the value of the given key in the persistent data store (null if the key hasn't
	//          been set yet).  If a set operation is being performed, null.
	my.DataStore = function(key, value) {
		if (value != null) {
			// Set
			my._dataStore[key] = value;
			return null;
		}
		else {
			// Get
			return my._dataStore[key];
		}
	};
	
	// Starts or stops a named timer.  Note that these timers are *global*, and will continue to run until the shutdown of the application
	// in which they were launched.  If, for example, you wish for a timer only to run as long as a particular window is open, you must
	// register a function to be executed on window close that will end the named timer you created for that window.
	// 
	// name: The name of the timer in question.
	// stopTimer: If true, stops the given timer.  Otherwise, starts the given timer.
	// interval: (optional) If starting a timer, specifies the interval in milliseconds between timer ticks.
	// callbackFn: (optional) If starting a timer, specifies the function to callback each timer tick.
	// dontRepeat: (optional) If starting a timer, this being true specifies that this timer should just tick once, and not repeat.
	// 
	// returns: null.
	my.Timer = function(name, stopTimer, interval, callbackFn, dontRepeat) {
		// Lazy init of named timers store
		if (my.DataStore("_timers") == null) {
			my.DataStore("_timers", {});
		}
		
		if (!stopTimer) {
			// Sanity check some arguments when starting a new timer
			if (interval == null) {
				my.ConsoleOutErr("When starting a timer, 'interval' argument must be non-null.", false);
				return null;
			}
			if (callbackFn == null || typeof(callbackFn) != "function") {
				my.ConsoleOutErr("When starting a timer, 'callbackFn' argument must be non-null, and must be a callback function.", false);
				return null;
			}
			
			// Start timer and store its data
			if (my.DataStore("_timers")[name] != null) {
				// The timer object may exist and yet the timer may have finished, which we can rely on detecting by the fact that it's
				// indicated by the 'callback' property being null; see:
				// http://mxr.mozilla.org/mozilla-central/source/xpcom/threads/nsTimerImpl.cpp#547
				// The callback reference is nullified when called, and then may be set back to what it was before if the timer is to keep
				// repeating; it's just the way it's implemented, apparently on purpose.
				if (my.DataStore("_timers")[name].callback == null) {
					my.DataStore("_timers")[name] = null;
				}
				else {
					my.ConsoleOutErr("Timer with name '" + name + "' already exists; cannot start another.", false);
					return null;
				}
			}
			
			var thisTimer =
				Cc["@mozilla.org/timer;1"]
				.createInstance(Ci.nsITimer);
			thisTimer.initWithCallback(callbackFn, interval, dontRepeat ? Ci.nsITimer.TYPE_ONE_SHOT : Ci.nsITimer.TYPE_REPEATING_SLACK);
			my.DataStore("_timers")[name] = thisTimer;
		}
		else {
			// Stop timer and remove its data
			thisTimer = my.DataStore("_timers")[name];
			if (thisTimer == null) {
				my.ConsoleOutErr("Couldn't find timer with name '" + name + "'.", false);
				return null;
			}
			
			thisTimer.cancel();
			
			my.DataStore("_timers")[name] = null;
		}
	};
	
	// Opens a new 'secondary' browser window and loads the specified resource in it.
	// 
	// theURL: The URL to the specified resource to open.
	// 
	// returns: null.
	my.OpenURL = function(theURL) {
		// Conveniently, there's a constantly-available singleton instance of a hidden nsIDOMWindow which is maintained by the
		// nsIAppShellService, during the lifetime of the application.  As each nsIDOMWindow provides an open() method, which 'creates
		// a new secondary browser window and loads the referenced resource', we can use this to open a new non-modal secondary window
		// when we want to open a new URL in the browser, even when this code is being called from within a modal dialog!
		
		var appShellService =
			Cc["@mozilla.org/appshell/appShellService;1"]
			.getService(Ci.nsIAppShellService);
		
		appShellService.hiddenDOMWindow.open(theURL);
	};
	
	// Gets an nsIPrefBranch object which points to the root pref branch.
	// 
	// returns: The nsIPrefBranch object.
	my.GetRootPrefBranch = function() {
		// Use my.xyz to force JS to use this function's container var's scope
		if (my._rootPrefBranch) { return my._rootPrefBranch; }
		else {
			var prefBranch =
				Cc["@mozilla.org/preferences-service;1"]
				.getService(Ci.nsIPrefBranch);
			
			return my._rootPrefBranch = prefBranch;
		}
	};
	
	// Gets the *default* pref branch with the specified name; pref values returned from this branch will be their defaults, even if
	// they have been set to something different by the user.
	// 
	// branchName: The name of the *default* pref branch to return.
	// 
	// returns: The *default* pref branch with the name specified.
	my.GetDefaultBranch = function(branchName) {
		var prefSvc =
			Cc["@mozilla.org/preferences-service;1"]
			.getService(Ci.nsIPrefService);
		
		return prefSvc.getDefaultBranch(branchName);
	};
	
	// Checks whether or not a preference is locked.
	// 
	// prefName: The name of the preference to check.
	// 
	// returns: If the pref is locked, true; otherwise false;
	my.CheckPrefIsLocked = function(prefName) {
		var rb = my.GetRootPrefBranch();
		return rb.prefIsLocked(prefName);
	};
	
	// Gets or sets a string preference.
	// 
	// options: An object containing data about the get/set operation to perform.
	// options.prefName: The name of the preference to get/set.
	// options.defaultVal: (optional) If getting a pref and the pref doesn't exist, this is returned instead of null.
	// options.doReset: (optional) If true, this resets the pref to its default value (or clears a user-defined pref entirely).  Overrides the options.setTo
	//                             option, so if this value is true, options.setTo is ignored.
	// options.setTo: (optional) If this is specified, sets the given pref to this value instead of getting its value.
	// options.mustExist: (optional) If this is true, a warning is sent to the error console if the pref to get/set doesn't exist.
	// options.isLocalized: (optional) If this is true, treats this as a complex localized string-type preference.
	// 
	// returns: If performing a set operation - null.
	//          If performing a get operation - if the pref exists and is of the right type, the value of the pref.
	//                                          if the pref exists and is of the wrong type, null.
	//                                          if the pref doesn't exist, null (or options.defaultVal if it was supplied).
	my.PrefString = function(options) {
		var rb = my.GetRootPrefBranch();
		var retVal = rb.getPrefType(options.prefName);
		
		var doGetOperation = (options.setTo == null && !options.doReset);
		if (retVal == rb.PREF_INVALID) {
			if (options.mustExist) {
				// Pref doesn't exist but must exist; this is an error
				my.ConsoleOutErr("Pref '" + options.prefName + "' must exist, but doesn't!  Cannot " + (doGetOperation ? "get" : "set") + " pref.", true);
				if (!doGetOperation) { return null; }
			}
			
			if (doGetOperation && options.defaultVal != null) { return options.defaultVal; }
			else if (doGetOperation) { return null; }
		}
		else if (retVal != rb.PREF_STRING) {
			// Pref exists but is not of string type; cannot get/set, this is an error
			my.ConsoleOutErr("Error " + (doGetOperation ? "getting" : "setting") + " string pref '" + options.prefName + "'; pref already exists but is not a string!", false);
			return null;
		}
		
		// Do get/set
		if (doGetOperation) {
			// Get pref
			if (options.isLocalized) {
				return rb.getComplexValue(options.prefName, Ci.nsIPrefLocalizedString).data;
			}
			else {
				return rb.getCharPref(options.prefName);
			}
		}
		else if (options.doReset) {
			// Reset pref
			rb.clearUserPref(options.prefName);
			return null;
		}
		else {
			// Set pref
			rb.setCharPref(options.prefName, options.setTo);
			return null;
		}
	};
	
	// Gets or sets an integer preference.
	// 
	// options: An object containing data about the get/set operation to perform.
	// options.prefName: The name of the preference to get/set.
	// options.defaultVal: (optional) If getting a pref and the pref doesn't exist, this is returned instead of null.
	// options.doReset: (optional) If true, this resets the pref to its default value (or clears a user-defined pref entirely).  Overrides the options.setTo
	//                             option, so if this value is true, options.setTo is ignored.
	// options.setTo: (optional) If this is specified, sets the given pref to this value instead of getting its value.
	// options.mustExist: (optional) If this is true, a warning is sent to the error console if the pref to get/set doesn't exist.
	// 
	// returns: If performing a set operation - null.
	//          If performing a get operation - if the pref exists and is of the right type, the value of the pref.
	//                                          if the pref exists and is of the wrong type, null.
	//                                          if the pref doesn't exist, null (or options.defaultVal if it was supplied).
	my.PrefInt = function(options) {
		var rb = my.GetRootPrefBranch();
		var retVal = rb.getPrefType(options.prefName);
		
		var doGetOperation = (options.setTo == null && !options.doReset);
		if (retVal == rb.PREF_INVALID) {
			if (options.mustExist) {
				// Pref doesn't exist but must exist; this is an error
				my.ConsoleOutErr("Pref '" + options.prefName + "' must exist, but doesn't!  Cannot " + (doGetOperation ? "get" : "set") + " pref.", true);
				if (!doGetOperation) { return null; }
			}
			
			if (doGetOperation && options.defaultVal != null) { return options.defaultVal; }
			else if (doGetOperation) { return null; }
		}
		else if (retVal != rb.PREF_INT) {
			// Pref exists but is not of integer type; cannot get/set, this is an error
			my.ConsoleOutErr("Error " + (doGetOperation ? "getting" : "setting") + " integer pref '" + options.prefName + "'; pref already exists but is not an integer!", false);
			return null;
		}
		
		// Do get/set
		if (doGetOperation) {
			// Get pref
			return rb.getIntPref(options.prefName);
		}
		else if (options.doReset) {
			// Reset pref
			rb.clearUserPref(options.prefName);
			return null;
		}
		else {
			// Set pref
			rb.setIntPref(options.prefName, options.setTo);
			return null;
		}
	};
	
	// Gets or sets a boolean preference.
	// 
	// options: An object containing data about the get/set operation to perform.
	// options.prefName: The name of the preference to get/set.
	// options.defaultVal: (optional) If getting a pref and the pref doesn't exist, this is returned instead of null.
	// options.doReset: (optional) If true, this resets the pref to its default value (or clears a user-defined pref entirely).  Overrides the options.setTo
	//                             option, so if this value is true, options.setTo is ignored.
	// options.setTo: (optional) If this is specified, sets the given pref to this value instead of getting its value.
	// options.mustExist: (optional) If this is true, a warning is sent to the error console if the pref to get/set doesn't exist.
	// 
	// returns: If performing a set operation - null.
	//          If performing a get operation - if the pref exists and is of the right type, the value of the pref.
	//                                          if the pref exists and is of the wrong type, null.
	//                                          if the pref doesn't exist, null (or options.defaultVal if it was supplied).
	my.PrefBool = function(options) {
		var rb = my.GetRootPrefBranch();
		var retVal = rb.getPrefType(options.prefName);
		
		var doGetOperation = (options.setTo == null && !options.doReset);
		if (retVal == rb.PREF_INVALID) {
			if (options.mustExist) {
				// Pref doesn't exist but must exist; this is an error
				my.ConsoleOutErr("Pref '" + options.prefName + "' must exist, but doesn't!  Cannot " + (doGetOperation ? "get" : "set") + " pref.", true);
				if (!doGetOperation) { return null; }
			}
			
			if (doGetOperation && options.defaultVal != null) { return options.defaultVal; }
			else if (doGetOperation) { return null; }
		}
		else if (retVal != rb.PREF_BOOL) {
			// Pref exists but is not of boolean type; cannot get/set, this is an error
			my.ConsoleOutErr("Error " + (doGetOperation ? "getting" : "setting") + " boolean pref '" + options.prefName + "'; pref already exists but is not a boolean!", false);
			return null;
		}
		
		// Do get/set
		if (doGetOperation) {
			// Get pref
			return rb.getBoolPref(options.prefName);
		}
		else if (options.doReset) {
			// Reset pref
			rb.clearUserPref(options.prefName);
			return null;
		}
		else {
			// Set pref
			rb.setBoolPref(options.prefName, options.setTo);
			return null;
		}
	};
	
	// Creates and returns a 'deep' DOM walker (walks anonymous DOM nodes as well as explicit DOM nodes) whose root node is the given node.
	// See for further discussion/info: http://groups.google.com/group/mozilla.dev.platform/browse_thread/thread/d89ed07aea746c50
	// 
	// rootNode: The root node to start DOM walking at.
	// showSubDocuments: (optional) If true, the walker will show sub-documents (eg. inside IFRAME); otherwise, it won't.
	// dontShowAnonymousContent: (optional) If true, the walker will NOT show anonymous content; otherwise, it will.
	// 
	// returns: The created 'deep' DOM walker.
	my.CreateDeepDOMWalker = function(rootNode, showSubDocuments, dontShowAnonymousContent) {
		// Create new deep DOM walker
		var walker =
			Cc["@mozilla.org/inspector/deep-tree-walker;1"]
			.createInstance(Ci.inIDeepTreeWalker);
		
		// Set its attributes
		walker.showAnonymousContent = (!dontShowAnonymousContent ? true : false);
		walker.showSubDocuments = (showSubDocuments == true ? true : false);
		walker.init(rootNode, Ci.nsIDOMNodeFilter.SHOW_ALL);
		
		// nb. nsIDOMNodeFilter should officially be documented at:
		//     https://developer.mozilla.org/en/XPCOM_Interface_Reference/nsIDOMNodeFilter
		//     ... as is:
		//     https://developer.mozilla.org/en/XPCOM_Interface_Reference/nsIDOMNode
		//     ... but at the time of writing, its official documentation page didn't exist, so try this unofficial one:
		//     http://www.oxymoronical.com/experiments/apidocs/interface/nsIDOMNodeFilter
		
		return walker;
	};
	
	// Checks to see whether a DOM node currently has a given class applied to it.
	// 
	// nodeToCheck: The DOM element node to check the class(es) of.
	// className: The name of the class to check for.
	// 
	// returns: If the given DOM node has the given class, true; otherwise false.
	my.NodeHasClass = function(nodeToCheck, className) {
		if (!nodeToCheck || !nodeToCheck.className) { return false; }
		
		var nodeClassNames = nodeToCheck.className.split(" ");
		for (var i=0; i < nodeClassNames.length; i++) {
			if (nodeClassNames[i] == className) { return true; }
		}
		
		return false;
	};
	
	// Finds the first DOM element node (or all nodes) whose list of classes contains the given class name, among the given node to search and all its
	// children, recursively.  Can check anonymous DOM nodes as well as explicit DOM nodes.  The class name to search for IS case-sensitive.
	// Basically, a 'getElementByClass()' equivalent for document.getElementById().
	// 
	// nodeToSearch: The DOM element node to search recursively.
	// classString: The case-sensitive name of the class to try and match.
	// findAll: If true, returns all matching element nodes found in an array; otherwise, just returns the first matching node.
	// checkSubDocuments: (optional) If true, will check sub-documents (eg. inside IFRAME) in the DOM tree for the given class; otherwise, it won't.
	// dontCheckAnonymousContent: (optional) If true, will NOT check anonymous content in the DOM tree for the given class; otherwise, it will.
	// 
	// returns: If findAll is false and an element whose classes list contains the given class name was found, the found DOM element node; otherwise null.
	//          If findAll is true and element(s) whose classes lists contain the given class name were found, an array of the element(s); otherwise an
	//          empty array.
	my.FindNodeByClass = function(nodeToSearch, classString, findAll, checkSubDocuments, dontCheckAnonymousContent) {
		return my.FindNodesByCriteria(nodeToSearch, 1, classString, findAll, checkSubDocuments, dontCheckAnonymousContent);
	};
	
	// Finds the first DOM element node (or all nodes) matching the given ID, among the given node to search and all its children, recursively.  Can check
	// anonymous DOM nodes as well as explicit DOM nodes.  The ID to search for IS case-sensitive.
	// 
	// nodeToSearch: The DOM element node to search recursively.
	// idString: The case-sensitive ID to try and match.
	// findAll: If true, returns all matching element nodes found in an array; otherwise, just returns the first matching node.
	// checkSubDocuments: (optional) If true, will check sub-documents (eg. inside IFRAME) in the DOM tree for the given ID; otherwise, it won't.
	// dontCheckAnonymousContent: (optional) If true, will NOT check anonymous content in the DOM tree for the given ID; otherwise, it will.
	// 
	// returns: If findAll is false and an element matching the given ID was found, the found DOM element node; otherwise null.
	//          If findAll is true and element(s) matching the given ID were found, an array of the element(s); otherwise an empty array.
	my.FindNodeById = function(nodeToSearch, idString, findAll, checkSubDocuments, dontCheckAnonymousContent) {
		return my.FindNodesByCriteria(nodeToSearch, 0, idString, findAll, checkSubDocuments, dontCheckAnonymousContent);
	};
	
	// Finds the first DOM element node (or all nodes) matching the given search criteria, among the given node to search and all its children, recursively.
	// Can check anonymous DOM nodes as well as explicit DOM nodes.
	// 
	// nodeToSearch: The DOM element node to search recursively.
	// matchType: 1: Matches by class.  Treats matchText as a case-sensitive class name string whose value must be included in the node's list of classes.
	//            2: Matches by local element name.  Treats matchText as a case-INSENSITIVE local element name string (eg. 'image') whose value must match
	//               the node's local element name.
	//            3: Matches by qualified element name.  Treats matchText as a case-INSENSITIVE qualified element name string (eg. 'xul:image') whose value
	//               must match the node's qualified element name.
	//            0 (or any other value): Matches by ID.  Treats matchText as a case-sensitive ID string whose value must match the node's ID.
	// matchText: The text to try and match against.
	// findAll: If true, returns all matching element nodes found in an array; otherwise, just returns the first matching node.
	// checkSubDocuments: (optional) If true, will check sub-documents (eg. inside IFRAME) in the DOM tree for the given criteria; otherwise, it won't.
	// dontCheckAnonymousContent: (optional) If true, will NOT check anonymous content in the DOM tree for the given criteria; otherwise, it will.
	// 
	// returns: If findAll is false and an element matching the given criteria was found, the found DOM element node; otherwise null.
	//          If findAll is true and element(s) matching the given criteria were found, an array of the elements(s); otherwise an empty array.
	my.FindNodesByCriteria = function(nodeToSearch, matchType, matchText, findAll, checkSubDocuments, dontCheckAnonymousContent) {
		if (!nodeToSearch || !matchText) { return null; }
		
		var foundArray = [];
		// Check the root node to search itself
		if (my.AddToArrayIfMatch(foundArray, nodeToSearch, matchType, matchText) && !findAll) { return foundArray[0]; }
		
		// Create 'deep' DOM walker for child nodes (checks anonymous nodes as well as explicit ones)
		var walker;
		if (!(walker = miscUtils.CreateDeepDOMWalker(nodeToSearch, checkSubDocuments, dontCheckAnonymousContent))) {
			miscUtils.ConsoleOutErr("Couldn't create deep DOM walker for node to search!");
			return null;
		}
		
		// Check child nodes recursively ('deep' DOM walker's .nextNode() method helpfully does the DOM walking recursively for us, checking
		// children's children, etc. in the correct order - basically in the order they'd appear in DOM Inspector).
		var currentChildNode;
		if (!(currentChildNode = walker.firstChild())) { return findAll ? foundArray : null; }
		
		do {
			if (my.AddToArrayIfMatch(foundArray, currentChildNode, matchType, matchText) && !findAll) { return foundArray[0]; }
		}
		while (currentChildNode = walker.nextNode());
		
		if (findAll) {
			// Return all found node(s)
			return foundArray;
		}
		else {
			// Child nodes checked recursively, and we've not returned yet?  Then nothing was found.
			return null;
		}
	};
	
	// Adds a given DOM element to a given array if it matches the given criteria.
	// 
	// array: The array to add the given DOM element to if it matches.
	// elementToMatch: The DOM element to try and match against the given criteria.
	// matchType: 1: Matches by class.  Treats matchText as a case-sensitive class name string whose value must be included in the node's list of classes.
	//            2: Matches by local element name.  Treats matchText as a case-INSENSITIVE local element name string (eg. 'image') whose value must match
	//               the node's local element name.
	//            3: Matches by qualified element name.  Treats matchText as a case-INSENSITIVE qualified element name string (eg. 'xul:image') whose value
	//               must match the node's qualified element name.
	//            0 (or any other value): Matches by ID.  Treats matchText as a case-sensitive ID string whose value must match the node's ID.
	// matchText: The text to try and match against.
	// 
	// returns: If element matched criteria and was therefore added to the array, true; otherwise false.
	my.AddToArrayIfMatch = function(array, elementToMatch, matchType, matchText) {
		if (!array) {
			miscUtils.ConsoleOutErr("No array passed in to add matched element to!");
			return null;
		}
		
		var matched = false;
		switch (matchType) {
			case 1:
				// Does element have given class?
				if (my.NodeHasClass(elementToMatch, matchText)) { matched = true; }
				break;
			case 2:
				// Does element local name match given local name?
				if (elementToMatch.localName == matchText) { matched = true; }
				break;
			case 3:
				// Does element qualified name match given qualified name?
				if (elementToMatch.nodeName == matchText) { matched = true; }
				break;
			case 0:
			default:
				// Does element ID match given ID?
				if (elementToMatch.id == matchText) { matched = true; }
				break;
		}
		
		if (matched) {
			array.push(elementToMatch);
			return true;
		}
		
		return false;
	};
	
	// Attempts to shutdown or restart the application.
	// 
	// doRestart: If true, attempts an application restart; otherwise, attempts an application shutdown.
	// 
	// returns: If the shutdown failed (eg. quit request was canceled by something), false; otherwise true.
	my.ShutdownApp = function(doRestart) {
		// Create instance of Javascript box for XPCOM boolean value, and get observer service
		var cancelQuit =
			Cc["@mozilla.org/supports-PRBool;1"]
			.createInstance(Ci.nsISupportsPRBool);
		var obsService =
			Cc["@mozilla.org/observer-service;1"]
			.getService(Ci.nsIObserverService);
		
		if (doRestart) {
			obsService.notifyObservers(cancelQuit, "quit-application-requested", "restart");
		}
		else {
			obsService.notifyObservers(cancelQuit, "quit-application-requested", null);
		}
		if (cancelQuit.data) {
			// Somebody canceled our quit request
			return false;
		}
		
		var startupService =
			Cc["@mozilla.org/toolkit/app-startup;1"]
			.getService(Ci.nsIAppStartup);
		if (doRestart) {
			startupService.quit(Ci.nsIAppStartup.eAttemptQuit | Ci.nsIAppStartup.eRestart);
		}
		else {
			startupService.quit(Ci.nsIAppStartup.eAttemptQuit);
		}
		return true;
	};
	
	// Prompts a user to click a button with a popup (modal to the given window), which contains a 'question mark' icon, and an optional checkbox with text.
	// 
	// window: The window that this popup prompt will be modal to.  Other browser windows will still be accessible when the popup is open.
	// title: The title of the popup box.
	// mainText: The main text of the popup box.
	// btn0Text: The text of the first button in the popup box (null for no button).
	// btn1Text: The text of the second button in the popup box (null for no button).
	// btn2Text: The text of the third button in the popup box (null for no button).
	// checkboxText: The text of the checkbox in the popup box (null for no checkbox+text).
	// checkboxValue: A Javascript object to set/receive the checked/unchecked value of the checkbox.  The value of the checkbox is determined by
	//                the 'value' property of this object, which should be set to 'true' if the checkbox is to be checked when the prompt displays,
	//                or 'false' otherwise.  Upon return, 'value' will be set to 'true' or 'false', depending on whether the checkbox was checked
	//                or not when the user dismissed the prompt dialog.
	//                Note that if checkboxText is null (and therefore no checkbox will be displayed), this can simply be null.
	// defaultBtn: Which of the buttons should be selected by default; should be 0, 1, or 2.  Defaults to 0.
	// 
	// returns: A value indicating which button was clicked; 0, 1, or 2 for button 0, button 1, or button 2 respectively.
	my.ModalPrompt = function(window, title, mainText, btn0Text, btn1Text, btn2Text, checkboxText, checkboxValue, defaultBtn) {
		// Get the Prompt Service, which is what's generally used for giving modal prompts to the user
		var promptService =
			Cc["@mozilla.org/embedcomp/prompt-service;1"]
			.getService(Ci.nsIPromptService);
		
		// Focus the window before prompting; this will raise any minimized window, which will make it obvious which window the prompt is for and will
		// solve the problem of windows "obscuring" the prompt.  See bug #350299 for more details.
		window.focus();
		
		var buttonFlags = 0;
		if (btn0Text != null) {
			buttonFlags += promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_0;
		}
		if (btn1Text != null) {
			buttonFlags += promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_1;
		}
		if (btn2Text != null) {
			buttonFlags += promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_2;
		}
		if (defaultBtn == null || defaultBtn < 0 || defaultBtn > 2) {
			defaultBtn = 0;
		}
		if (defaultBtn == 0) {
			buttonFlags += promptService.BUTTON_POS_0_DEFAULT;
		}
		else if (defaultBtn == 1) {
			buttonFlags += promptService.BUTTON_POS_1_DEFAULT;
		}
		else if (defaultBtn == 2) {
			buttonFlags += promptService.BUTTON_POS_2_DEFAULT;
		}
		
		if (checkboxValue == null) {
			checkboxValue = { value: true };
		}
		var btnPressed = promptService.confirmEx(
			window,
			title,
			mainText,
			buttonFlags,
			btn0Text,
			btn1Text,
			btn2Text,
			checkboxText,
			checkboxValue
		);
		
		return btnPressed;
	};
	
	// Prompts a user to click OK or Cancel, displaying a given message (modal to the given window), along with a 'question mark' icon.
	// 
	// window: The window that this popup prompt will be modal to.  Other browser windows will still be accessible when the popup is open.
	// title: The title of the popup box.
	// messageText: The message to display as the prompt of the popup box.
	// 
	// returns: A value indicating which button was clicked; true for OK, false for Cancel or just closing the prompt.
	my.ModalConfirm = function(window, title, messageText) {
		// Get the Prompt Service, which is what's generally used for giving modal prompts to the user
		var promptService =
			Cc["@mozilla.org/embedcomp/prompt-service;1"]
			.getService(Ci.nsIPromptService);
		
		// Focus the window before prompting; this will raise any minimized window, which will make it obvious which window the prompt is for and will
		// solve the problem of windows "obscuring" the prompt.  See bug #350299 for more details.
		window.focus();
		
		var btnPressed = promptService.confirm(
			window,
			title,
			messageText
		);
		
		return btnPressed;
	};
	
	// Alias for GetJsVarData.
	my.GetJsVarInfo = function(jsVar, jsVarName) {
		return my.GetJsVarData(jsVar, jsVarName);
	};
	
	// Generates and returns a string giving debug data on a given Javascript variable.
	// 
	// jsVar: The Javascript variable to generate data on.
	// jsVarName: (optional) If specified, this will be used as the 'name' of the variable passed in.  Otherwise, it will default to 'unnamed'.
	// 
	// returns: The string of debug data generated for the given Javascript variable.
	my.GetJsVarData = function(jsVar, jsVarName) {
		if (jsVarName == null) { jsVarName = "(unnamed)"; }
		var isObject = false;
		
		try {
			// Because of the fix to: https://bugzilla.mozilla.org/show_bug.cgi?id=254067
			// ... an 'Object' from one window can be different to an 'Object' from another, or a JSM.  We need to see whether the jsVar is
			// an instance of ITS global scope's Object, not ours.
			if (jsVar instanceof Components.utils.getGlobalForObject(jsVar).Object) { isObject = true; }
		}
		catch (ex) {
			// If we get an exception here, we're probably dealing with a non-Object and so .getGlobalForObject(nonObject) fails.
			isObject = false;
		}
		
		var varData = "";
		if (isObject) {
			varData += "Javascript object\n";
			varData += "=================\n";
		}
		else {
			varData += "Javascript variable\n";
			varData += "===================\n";
		}
		varData += "- " + jsVarName + " [" + typeof(jsVar) + "]: " + jsVar + "\n";
		if (isObject) {
			varData += "\n";
			varData += "Object's properties\n";
			varData += "===================\n";
			var count = 0;
			for (var key in jsVar) {
				var strTypeName = "";
				try {
					strTypeName = typeof(jsVar[key]);
				}
				catch (ex) {
					strTypeName = "???";
				}
				var propertyData = "- " + key + " [" + strTypeName + "]";
				
				try {
					if (jsVar[key] == null) {
						propertyData += ", is null";
					}
					else {
						propertyData += ": " + jsVar[key];
					}
				}
				catch (ex) {
					// If we get an exception here, we couldn't access this object property for some reason.
					propertyData += ", (couldn't access property: " + ex + ")";
				}
				
				varData += propertyData + "\n";
				count++;
			}
			if (count == 0) {
				varData += "(none)\n";
			}
		}
		
		return varData;
	};
	
	// Sets a XUL textbox to the given new value.  Optionally takes a callback function to be called if the value has changed.
	// 
	// tbox: The XUL textbox object.
	// newVal: The new value to set the value of the textbox to.
	// onChange: (optional) If newVal was different to the existing value of the textbox, this callback function is called.  The textbox
	//                      object, and the old and new values, are passed to the function.
	// 
	// returns: null.
	my.SetTextbox = function(tbox, newVal, onChange) {
		var doCallback = false;
		var oldVal = tbox.value;
		
		if (onChange != null) {
			if (typeof(onChange) != "function") {
				my.ConsoleOutErr("onChange param must be a callback function!");
				return;
			}
			
			if (oldVal != newVal) {
				doCallback = true;
			}
		}
		
		if (oldVal != newVal) { tbox.value = newVal; }
		
		if (doCallback) {
			onChange(tbox, oldVal, newVal);
		}
	};
	
	// Sets a XUL radio group selection to the given new value.  Optionally takes a callback function to be called if the selection has changed.
	// 
	// rgrp: The XUL radio group object.
	// newVal: The new value to set the selection of the radio group to.
	// onChange: (optional) If newVal was different to the existing selection of the radio group, this callback function is called.  The radio
	//                      group object, and the old and new values, are passed to the function.
	// 
	// returns: null.
	my.SetRadio = function(rgrp, newVal, onChange) {
		var doCallback = false;
		var oldVal = rgrp.value;
		
		if (onChange != null) {
			if (typeof(onChange) != "function") {
				my.ConsoleOutErr("onChange param must be a callback function!");
				return;
			}
			
			if (oldVal != newVal) {
				doCallback = true;
			}
		}
		
		if (oldVal != newVal) { rgrp.value = newVal; }
		
		if (doCallback) {
			onChange(rgrp, oldVal, newVal);
		}
	};
	
	// Attempts to find a stylesheet by href given an array of stylesheets through which to search, and the href of the desired stylesheet.  Searches
	// recursively through 'child' stylesheets as well (ones which have been imported from the stylesheets passed in) so that if the desired stylesheet
	// is anywhere in the 'tree' of stylesheets, it will be returned.
	// 
	// styleSheets: An array of stylesheets through which to search recursively for the desired stylesheet.
	// href: The href of the desired stylesheet.
	// debugOutput: (optional) If true, will output to console a list of the href's of all the stylesheets being searched through.
	// 
	// returns: If the desired stylesheet with an href matching the requested href was found, the stylesheet object; otherwise null.
	my.FindStyleSheet = function(styleSheets, href, debugOutput) {
		var hrefList;
		
		for each (var styleSheet in styleSheets) {
			if (debugOutput) {
				my.ConsoleOutErr("Searching stylesheet: " + styleSheet.href, true);
			}
			
			// Match requested href in this sheet?
			if (styleSheet.href == href) {
				return styleSheet;
			}
			
			// Match requested href in a child sheet?
			var styleSheetCssRules;
			try {
				styleSheetCssRules = styleSheet.cssRules;
			}
			catch (ex) {
				// Just trying to access a styleSheet's cssRules can fail horribly sometimes; on an exception, give up.
				styleSheetCssRules = null;
			}
			if (styleSheetCssRules != null) {
				var count=0;
				for each (var rule in styleSheet.cssRules) {
					if (rule.type == rule.IMPORT_RULE && rule.styleSheet != null) {
						var foundSheet = my.FindStyleSheet([ rule.styleSheet ], href, debugOutput);
						if (foundSheet) {
							return foundSheet;
						}
					}
				}
			}
		}
		
		// Nothing found
		return null;
	};
	
	// Gets comprehensive CSS style rules for the given DOM object.  Note that this returns an object with a generic array interface intended for internal
	// use (nsISupportsArray), and as such it may be tricky to use properly.  See the following for the interface definition:
	// http://mxr.mozilla.org/mozilla-central/source/xpcom/ds/nsISupportsArray.idl#34
	// 
	// domObject: The DOM object to get comprehensive CSS style rules for.
	// 
	// returns: Comprehensive array of CSS style rules for the given DOM object.
	my.GetCssStyleRules = function(domObject) {
		var domUtils =
			Cc["@mozilla.org/inspector/dom-utils;1"]
			.getService(Ci.inIDOMUtils);
		
		return domUtils.getCSSStyleRules(domObject);
	};
	
	// Tests whether or not a given value is null/undefined.
	// 
	// toCheck: The value to check.
	// representAsString: (optional) If true, returns the result of the check represented as a string; otherwise, just returns true or false values.
	// 
	// returns: The result of checking whether the given value is null/undefined.
	my.IsNullOrUndefined = function(toCheck, representAsString) {
		if (toCheck === null) {
			return representAsString ? "(null)" : true;
		}
		if (toCheck === undefined) {
			return representAsString ? "(undef)" : true;
		}
		return representAsString ? "IS defined" : false;
	};
	
	// Writes the given text out to the file with the given filename, asynchronously.  Be aware that if this function is called many times in quick
	// succession on the same file (such as in a for loop appending to the same file), the output may not be linear as there will be a race condition
	// as to the order in which the different calls get to write their data out to the file.
	// This function should therefore be called on any one file once, with a significant delay after the call to allow time for the asynchronous
	// write to complete.
	// 
	// specialLocation: One of a number of strings specifying the special location that will be the base directory that the file will be written to.
	// filePathArray: An array of path components specifying the path *within the special location* to write the file to; the final array item must
	//                be the leaf node of the path: the filename itself.
	// textToWrite: The text to write to the file.
	// doAppend: (optional) If true, will append the text if the specified file already exists.
	// errorObject: (optional) If passed in, the "errorMsg" property on this object will be set to the error message if writing failed.
	// 
	// returns: If writing succeeded, true; otherwise false.
	my.WriteToFileAsync = function(specialLocation, filePathArray, textToWrite, doAppend, errorObject) {
		// Here are some of the values that can be specified for specialLocation, and the location it will map to:
		// 
		// STRING        MEANING
		// =====================
		// AChrom        %CurProcD%/chrome
		// APlugns       %CurProcD%/plugins
		// ComsD         %CurProcD%/components
		// CurProcD      Current working directory (usually the application's installation directory).
		// DefProfRt     The root of the default profile directory (for example /root/.mozilla).
		// DefRt         %CurProcD%/defaults
		// Desk          The user's desktop directory (for example ~/Desktop on Linux or Mac OS X, C:\Documents and Settings\username\Desktop on Windows).
		// DfltDwnld     The default Downloads directory (for example, ~/Downloads on Mac OS X).
		// Home          The user's home directory (for example, /home/username).
		// PrfDef        %installation%/defaults/pref
		// ProfD         The profile directory.
		// ProfDefNoLoc  %installation%/defaults/profile
		// ProfLD        Local Settings on Windows; where the network cache and fastload files are stored.
		// Progs         User Start menu programs directory (for example, C:\Documents and Settings\username\Start Menu\Programs). This is a Windows-specific value.
		// TmpD          The operating system's temporary files directory (for example, /tmp on Mac OS X and Linux).
		// UChrm         The user chrome directory in their profile: %profile%/chrome.
		// resource:app  The application directory in a XULRunner application.
		
		try {
			var file = my.jsm.FileUtils.getFile(specialLocation, filePathArray);
			var modeflags = my.jsm.FileUtils.MODE_WRONLY | my.jsm.FileUtils.MODE_CREATE;
			if (doAppend) {
				modeflags |= my.jsm.FileUtils.MODE_APPEND;
			}
			else {
				modeflags |= my.jsm.FileUtils.MODE_TRUNCATE;
			}
			
			var outputstream = my.jsm.FileUtils.openFileOutputStream(file, modeflags);
			var converter =
				Cc["@mozilla.org/intl/scriptableunicodeconverter"]
				.createInstance(Ci.nsIScriptableUnicodeConverter);
			converter.charset = "UTF-8";
			var inputstream = converter.convertToInputStream(textToWrite);
			
			// Note that because this is async, the write may still fail even if this function returns true; we have merely *started* the write successfully.
			my.jsm.NetUtil.asyncCopy(inputstream, outputstream);
		}
		catch (ex) {
			my.Dump("WriteToFile failed: " + ex);
			if (errorObject && typeof errorObject === 'object') {
				errorObject.errorMsg = ex;
			}
			return false;
		}
		
		return true;
	};
	
	// Gets the 'pessimistic' actual height of a standard dialog (containing a box for 'content', and a box for 'buttons' such as OK/Cancel) based on
	// the boxObject heights of the aforementioned two boxes.  This is based on testing in-the-wild of the difference between the boxObject height of
	// these element combined, and the actual on-screen height of the dialog box that is rendered to contain them, across various different platforms.
	// 
	// contentBox: The dialog's content box.
	// buttonBox: The dialog's button box.
	// 
	// returns: The 'pessimistic' actual height of the standard dialog containing the given boxes.
	my.GetPessimisticActualDialogHeight = function(contentBox, buttonBox) {
		// Calculate our pessimistically high estimate of what the actual dialog height might be.  So far, the biggest discrepancy we've found is with
		// Windows 7, where the actual dialog height is just under ~60px taller than the content and button box heights combined, so we add 60.
		var pessimisticActualDialogHeight = (contentBox.boxObject.height+buttonBox.boxObject.height) + 60;
		
		return pessimisticActualDialogHeight;
	};
	
	// *** Finally ***
	// Return |my| - everything we've assigned to it will be exported to the public var we're defining.
	return my;
	
}((miscUtils || {}), this));
