"use strict";
/*
    This file is part of the Pale Moon Locale Switcher extension.
    
    Copyright (C) 2014,2015 Jeremy Morton
    Copyright (C) 2016 Mark Straver BASc

    The Pale Moon Locale Switcher extension is free software: you can
    redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either
    version 3 of the License, or (at your option) any later version.

    The Pale Moon Locale Switcher extension is distributed in the hope that
    it will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
// Javascript Code Module - Web content language functionality
// Contains functionality for accessing the user's preferred web content language(s).

var EXPORTED_SYMBOLS = ["webContentLang"];

// Components shortcuts
const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;

var webContentLang = (function(my, moduleGlobalScope) {
	
	// Init - RUN FIRST TO INJECT MODULE DEPENDENCIES
	// ==============================================
	
	// Basically a fake constructor for dependency injection - should be run before any other functions in this lib
	my.Init = function(bundleAccepted, bundleRegions, bundleLanguages, bundlePreferences) {
		// bundleAccepted should be a string bundle pointing to a "locales accepted" properties file
		// bundleRegions should be a string bundle pointing to a "region descriptions" properties file
		// bundleLanguages should be a string bundle pointing to a "language descriptions" properties file
		// bundlePreferences should be a string bundle pointing to a "language coding preferences" properties file
		my._bundleAccepted = bundleAccepted;
		my._bundleRegions = bundleRegions;
		my._bundleLanguages = bundleLanguages;
		my._bundlePreferences = bundlePreferences;
	};
	
	
	
	// Private members (please don't use them directly from calling code)
	// ==================================================================
	
	my._appName = null;
	my._bundleAccepted = null;
	my._bundleRegions = null;
	my._bundleLanguages = null;
	my._bundlePreferences = null;
	my._availableLanguagesCache = null;

	// Language information container class
	function _LanguageInfo(aName, aABCD, aIsVisible) {
		this.name = aName;
		this.abcd = aABCD;
		this.isVisible = aIsVisible;
	};
	
	my._loadAvailableLanguages = function() {
		// Lazy-load available languages if necessary
		if (my._availableLanguagesCache == null) {
			my._availableLanguagesCache = [];
			
			// This is a parser for: resource://gre/res/language.properties
			// The file is formatted like so:
			// ab[-cd].accept=true|false
			//  ab = language
			//  cd = region
			
			// Read the available languages out of language.properties
			var strings = my._bundleAccepted.getSimpleEnumeration();
			while (strings.hasMoreElements()) {
				var currString = strings.getNext();
				if (!(currString instanceof Ci.nsIPropertyElement)) { break; }
				
				var property = currString.key.split(".");  // ab[-cd].accept
				if (property[1] == "accept") {
					var abCD = property[0];
					var abCDPairs = abCD.split("-");       // ab[-cd]
					var useABCDFormat = abCDPairs.length > 1;
					var ab = useABCDFormat ? abCDPairs[0] : abCD;
					var cd = useABCDFormat ? abCDPairs[1] : "";
					if (ab) {
						var language = my._bundleLanguages.GetStringFromName(ab);
						
						var region = "";
						if (useABCDFormat) {
							region = my._bundleRegions.GetStringFromName(cd);
						}
						
						var name = "";
						if (useABCDFormat) {
							var paramList = [language, region, abCD];
							name = my._bundlePreferences.formatStringFromName("languageRegionCodeFormat", paramList, paramList.length);
						}
						else {
							var paramList = [language, abCD];
							name = my._bundlePreferences.formatStringFromName("languageCodeFormat", paramList, paramList.length);
						}
						
						if (name && abCD) {
							my._availableLanguagesCache.push(new _LanguageInfo(name, abCD, currString.value == "true"));
						}
					}
				}
			}
			
			// Sort the list of languages by name
			my._availableLanguagesCache.sort(function(a, b) {
				return a.name.localeCompare(b.name);
			});
		}
	};
	
	my._getLanguageName = function(abcd) {
		my._loadAvailableLanguages();
		
		for (var i=0; i < this._availableLanguagesCache.length; i++) {
			if (abcd == this._availableLanguagesCache[i].abcd) {
				return this._availableLanguagesCache[i].name;
			}
		}
		
		return "";
	};
	
	
	
	// Public members
	// ==============
	
	// Gets an array of languages available to select from for preferred web content.
	// 
	// returns: An array of languages available to select from for preferred web content.
	my.GetLanguages = function() {
		my._loadAvailableLanguages();
		
		return my._availableLanguagesCache;
	};
	
	// Gets an array of languages contained in a comma-delimited string of language codes.
	// 
	// codesString: A comma-delimited string of language codes.
	// 
	// returns: An array of languages represented by the given string.
	my.GetLanguagesFromCodesString = function(codesString) {
		var languages = [];
		var languageCodes = codesString.toLowerCase().split(',');
		for (var i=0; i < languageCodes.length; i++) {
			languageCodes[i] = languageCodes[i].trim();
			var name = my._getLanguageName(languageCodes[i]);
			if (!name) { name = "[" + languageCodes[i] + "]"; }
			languages.push(new _LanguageInfo(name, languageCodes[i], true));
		}
		
		return languages.length == 1 && languages[0].name == "[]" ? [] : languages;
	};
	
	// Gets a comma-delimited string of language codes from an array of XUL listitem's representing languages.
	// 
	// languageItems: An array of XUL listitem's representing languages.
	// 
	// returns: A comma-delimited string of language codes corresponding to the given array of XUL listitem's.
	my.GetCodesStringFromLanguages = function(languageItems) {
		var languagesString = "";
		for (var i=0; i < languageItems.length; i++) {
			languagesString += ((i == 0 ? "" : ",") + languageItems[i].id);
		}
		
		return languagesString;
	};
	
	// *** Finally ***
	// Return |my| - everything we've assigned to it will be exported to the public var we're defining.
	return my;
	
}((webContentLang || {}), this));
